import functools
import logging
import logging.config
import os
import sys

import structlog
from structlog.processors import JSONRenderer

from . import util


def stringify_dict_keys(obj):
    if isinstance(obj, list):
        obj = [stringify_dict_keys(item) for item in obj]
    elif isinstance(obj, tuple):
        obj = tuple(stringify_dict_keys(item) for item in obj)
    elif isinstance(obj, dict):
        obj = {str(key): stringify_dict_keys(value) for key, value in obj.items()}
    return obj


try:
    import orjson

    json_dumps_inner = functools.partial(
        orjson.dumps, option=orjson.OPT_NAIVE_UTC | orjson.OPT_NON_STR_KEYS
    )
    json_dumps = lambda *arg, **kwargs: json_dumps_inner(*arg, **kwargs).decode("utf-8")
except ImportError:
    import json

    json_dumps = json.dumps


LOGGING_LEVEL_NAMES = list(
    map(
        logging.getLevelName,
        sorted(
            (
                logging.NOTSET,
                logging.DEBUG,
                logging.INFO,
                logging.WARN,
                logging.ERROR,
                logging.CRITICAL,
            )
        ),
    )
)
DEFAULT_LOGGING_LEVEL = logging.getLevelName(logging.WARNING)
PID = os.getpid()


def setup_logging(
    *,
    tty=sys.stdout.isatty(),
    level=logging.DEBUG,
    user_config=None,
    capture_warnings=True,
    redirect_print=False,
    json_renderer=JSONRenderer(
        serializer=lambda obj, **kwargs: json_dumps(stringify_dict_keys(obj), **kwargs)
    ),
):

    if isinstance(level, str):
        level = logging.getLevelName(level.upper())

    renderer = structlog.dev.ConsoleRenderer() if tty else json_renderer
    timestamper = structlog.processors.TimeStamper(fmt="ISO", utc=True)
    pre_chain = [
        # Add the log level and a timestamp to the event_dict if the log entry
        # is not from structlog.
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.processors.format_exc_info,
        timestamper,
        add_os_pid,
        # ExtractLogExtra(
        #     "spec",
        #     "url",
        #     "mimetype",
        #     "has_body",
        #     "swagger_yaml",
        #     "method",
        #     "path",
        #     "operation_id",
        #     "data",
        # ),
    ]

    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "structured": {
                "()": structlog.stdlib.ProcessorFormatter,
                "processor": renderer,
                "foreign_pre_chain": pre_chain,
            }
        },
        "handlers": {
            "default": {"class": "logging.StreamHandler", "formatter": "structured"}
        },
        "loggers": {"": {"handlers": ["default"], "level": level, "propagate": True}},
    }
    if user_config:
        util.merge_dict(user_config, dest=config)
    logging.config.dictConfig(config)

    logging.captureWarnings(capture_warnings)
    processors = [
        structlog.stdlib.add_log_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.PositionalArgumentsFormatter(),
        timestamper,
        add_os_pid,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
    ]

    structlog.configure(
        processors=processors,
        context_class=structlog.threadlocal.wrap_dict(dict),
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    if redirect_print:
        # redirect stdio print
        print_log = structlog.get_logger("print")
        sys.stderr = StdioToLog(print_log)
        sys.stdout = StdioToLog(print_log)

    # log uncaught exceptions
    sys.excepthook = uncaught_exception


class ExtractLogExtra:  # noqa: R0903

    """Extract log record attributes to structlog event_dict."""

    def __init__(self, *attrs):
        self.attrs = attrs

    def __call__(self, logger, method_name, event_dict):
        """
        Add the logger name to the event dict.
        """
        record = event_dict.get("_record")
        for attr_name in self.attrs:
            if hasattr(record, attr_name):
                attr = getattr(record, attr_name)
                event_dict[attr_name] = attr
        return event_dict


class StdioToLog:

    """Delegate sys.stdout to a logger."""

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def write(self, buf):  # noqa
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())

    def flush(self):  # noqa
        pass


def add_os_pid(logger, method_name, event_dict):  # noqa
    event_dict["pid"] = PID
    return event_dict


def uncaught_exception(ex_type, ex_value, tb):  # noqa: C0103
    log_ = structlog.get_logger("sys.excepthook")
    log_.critical(event="uncaught exception", exc_info=(ex_type, ex_value, tb))



import functools
import os
import sys
import typing

import attr
import cattr
import structlog
import tomlkit
import typing_inspect

from . import di, util

CNF_KEY = "buvar_config"


logger = structlog.get_logger()


# since we only need a single instance, we just hide the class magically
@functools.partial(lambda x: x())
class missing:
    def __repr__(self):
        return self.__class__.__name__


@attr.s(auto_attribs=True)
class ConfigValue:
    name: typing.Optional[str] = None
    help: typing.Optional[str] = None


def var(
    default=missing,
    converter=None,
    factory=missing,
    name=None,
    validator=None,
    help=None,  # noqa: W0622,
):
    return attr.ib(
        metadata={CNF_KEY: ConfigValue(name, help)},
        converter=converter,
        validator=validator,
        **({"default": default} if factory is missing else {"factory": factory}),
    )


def _env_to_bool(val):
    """
    Convert *val* to a bool if it's not a bool in the first place.
    """
    if isinstance(val, bool):
        return val
    val = val.strip().lower()
    if val in ("1", "true", "yes", "on"):
        return True

    return False


def _env_to_list(val):
    """Take a comma separated string and split it."""
    if isinstance(val, str):
        val = map(lambda x: x.strip(), val.split(","))
    return val


def bool_var(default=missing, name=None, help=None):  # noqa: W0622
    return var(default=default, name=name, converter=_env_to_bool, help=help)


def list_var(default=missing, name=None, help=None):  # noqa: W0622
    return var(default=default, name=name, converter=_env_to_list, help=help)


class ConfigSource(dict):

    """Config dict, with loadable sections.

    >>> @attr.s(auto_attribs=True)
    ... class FooConfig:
    ...     bar: float
    >>>
    >>>
    >>> source = {'foo': {'bar': 1.23}}
    >>> config = ConfigSource(source)
    >>> foo = config.load(FooConfig, "foo")
    >>> foo
    FooConfig(bar=1.23)
    """

    __slots__ = ("env_prefix",)

    def __init__(self, *sources, env_prefix: typing.Optional[str] = None):
        super().__init__()
        util.merge_dict(*sources, dest=self)
        # config = schematize(__source, cls, env_prefix=env_prefix)

        self.env_prefix: typing.Tuple[str, ...] = (env_prefix,) if env_prefix else ()

    def merge(self, *sources):
        util.merge_dict(*sources, dest=self)

    def load(self, config_cls, name=None):
        if name is None:
            values = self
        else:
            values = self.get(name, {})

        env_config = create_env_config(
            config_cls, *(self.env_prefix + (name,) if name else ())
        )
        values = util.merge_dict(values, env_config)
        # merge environment

        config = relaxed_converter.structure(values, config_cls)
        return config


class ConfigError(Exception):
    ...


# to get typing_inspect.is_generic_type()
ConfigType = typing.TypeVar("ConfigType", bound="Config")

skip_section = type.__new__(
    type, "skip_section", (object,), {"__repr__": lambda self: self.__class__.__name__}
)()


class Config:
    __buvar_config_section__: typing.Optional[str] = skip_section
    __buvar_config_sections__: typing.Dict[str, type] = {}

    def __init_subclass__(cls, section=skip_section, **kwargs):
        if section is skip_section:
            return
        if section in cls.__buvar_config_sections__:
            raise ConfigError(
                f"Config section `{section}` already defined!",
                cls.__buvar_config_sections__,
                cls,
            )
        cls.__buvar_config_section__ = section
        cls.__buvar_config_sections__[section] = cls

    @classmethod
    async def adapt(cls: typing.Type[ConfigType], source: ConfigSource) -> ConfigType:
        config = source.load(cls, cls.__buvar_config_section__)
        return config


def traverse_attrs(cls, *, target=None, get_type_hints=typing.get_type_hints):
    """Traverse a nested attrs structure, create a dictionary for each nested
    attrs class and yield all fields resp. path, type and target dictionary."""
    stack = [
        (
            target if target is not None else {},
            (),
            list(attr.fields(cls)),
            get_type_hints(cls),
        )
    ]
    while stack:
        target, path, fields, hints = stack.pop()
        while fields:
            field = fields.pop()
            field_path = path + (field.name,)
            field_type = hints[field.name]
            if attr.has(field_type):
                target[field.name] = field_target = {}
                # XXX should we yield also attrs classes?
                yield field_path, field_type, target

                stack.append((target, path, fields, hints))
                target, path, fields, hints = (
                    field_target,
                    field_path,
                    list(attr.fields(field_type)),
                    get_type_hints(field_type),
                )
            else:
                yield field_path, field_type, target


def create_env_config(cls, *env_prefix):
    frame = sys._getframe(1)
    get_type_hints = functools.partial(
        typing.get_type_hints, globalns=frame.f_globals, localns=frame.f_locals
    )

    env_config = {}
    for path, _, target in traverse_attrs(
        cls, target=env_config, get_type_hints=get_type_hints
    ):
        env_name = "_".join(map(lambda x: x.upper(), env_prefix + path))
        if env_name in os.environ:
            target[path[-1]] = os.environ[env_name]
    return env_config


class RelaxedConverter(cattr.Converter):
    """This py:obj:`RelaxedConverter` is ignoring undefined and defaulting to
    None on optional attributes."""

    def structure_attrs_fromdict(
        self, obj: typing.Mapping, cl: typing.Type
    ) -> typing.Any:
        """Instantiate an attrs class from a mapping (dict)."""
        conv_obj = {}
        dispatch = self._structure_func.dispatch
        for a in attr.fields(cl):
            # We detect the type by metadata.
            type_ = a.type
            if type_ is None:
                # No type.
                continue
            name = a.name
            try:
                val = obj[name]
            except KeyError:
                if typing_inspect.is_optional_type(type_):
                    if a.default in (missing, attr.NOTHING):
                        val = None
                    else:
                        val = a.default
                elif a.default in (missing, attr.NOTHING):
                    raise ValueError("Attribute is missing", a.name)
                else:
                    continue

            if a.converter is None:
                val = dispatch(type_)(val, type_)

            conv_obj[name] = val

        return cl(**conv_obj)


relaxed_converter = RelaxedConverter()


def generate_env_help(cls, env_prefix=""):
    """Generate a list of all environment options."""

    help = "\n".join(  # noqa: W0622
        "_".join((env_prefix,) + path if env_prefix else path).upper()
        for path, type, _ in traverse_attrs(cls)
        if not attr.has(type)
    )
    return help


def generate_toml_help(config_cls, *, parent=None):
    if parent is None:
        parent = tomlkit.table()
        doclines = trim(config_cls.__doc__).split("\n")
        for line in doclines:
            parent.add(tomlkit.comment(line))
        parent.add(tomlkit.nl())

    for attrib in attr.fields(config_cls):
        meta = attrib.metadata.get(CNF_KEY)
        if attr.has(attrib.type):
            # yield (attrib.name,), attrib
            sub_doc = generate_toml_help(attrib.type)
            parent.add(attrib.name, sub_doc)
        else:
            if meta:
                parent.add(tomlkit.comment(meta.help))

            if attrib.default in (missing, attr.NOTHING):
                parent.add(tomlkit.comment(f"{attrib.name} ="))
            else:
                default = (
                    attrib.default() if callable(attrib.default) else attrib.default
                )
                parent.add(attrib.name, default)

            parent.add(tomlkit.nl())

    return parent


def trim(docstring):
    # https://www.python.org/dev/peps/pep-0257/
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    return "\n".join(trimmed)


async def prepare():
    di.register(Config.adapt)



# cython: language_level=3
from ..components.c_components cimport Components, ComponentLookupError
from buvar import context


missing = object()


cdef class ResolveError(Exception):
    pass


cdef _get_name_or_default(cmps, target, name=None):
    # find in components
    if name is not None:
        try:
            component = cmps.get(target, name=name)
            return component
        except ComponentLookupError:
            pass

    component = cmps.get(target, name=None)
    return component


cdef prepare_components(tuple targets, dict dependencies):
    # create components
    cdef Components cmps = Components()
    cdef list stack

    # add default unnamed dependencies
    # every non-default argument of the same type gets its value
    # XXX is this good?
    for name, dep in dependencies.items():
        cmps.add(dep)

    cdef Components current_context = context.current_context()

    # add current context
    if not current_context:
        stack = []
    else:
        stack = current_context.stack
    cmps = cmps.push(*stack)

    # add default named dependencies
    cmps = cmps.push()
    for name, dep in dependencies.items():
        cmps.add(dep, name=name)

    return cmps


cdef class AdaptersImpl:
    async def nject(self, *targets, **dependencies):
        """Resolve all dependencies and return the created component."""

        cdef tuple _targets = targets

         # create components
        cdef Components cmps = prepare_components(_targets, dependencies)        # create components
        # find the proper components to instantiate that class
        cdef list injected = [
            await self.resolve_adapter(cmps, target, name=name)
            for name, target in ((None, target) for target in targets)
        ]
        if len(targets) == 1:
            return injected[0]
        return injected

    def get_possible_adapters(self, target):
        for lookup in self.lookups:
            yield from lookup(self, target)

    async def resolve_adapter(self, cmps, target, *, name=None, default=missing):
        # find in components
        try:
            component = _get_name_or_default(cmps, target, name)
            return component
        except ComponentLookupError:
            pass

        cdef list resolve_errors = []
        possible_adapters = self.get_possible_adapters(target)
        cdef dict adapter_args

        adptr = None
        for adptr in possible_adapters:
            try:
                adapter_args = {
                    name: await self.resolve_adapter(
                        cmps, dependency_target, name=name, default=default
                    )
                    for name, dependency_target, default in adptr.args
                }
            except ResolveError as ex:
                # try next adapter
                resolve_errors.append(ex)
            else:
                component = await adptr.create(target, **adapter_args)
                # we do not use the name
                cmps.add(component)
                return component

        if default is not missing:
            return default

        # have we tried at least one adapter
        if adptr is None:
            raise ResolveError("No possible adapter found", target, [])
        raise ResolveError("No adapter dependencies found", target, resolve_errors)



import socket
import typing

import aiohttp.web
import attr
from buvar import plugin, config, context
import structlog

try:
    from ssl import SSLContext
except ImportError:  # pragma: no cover
    SSLContext = typing.Any  # type: ignore


@attr.s(auto_attribs=True)
class AioHttpConfig:
    host: typing.Optional[str] = None
    port: typing.Optional[int] = None
    path: typing.Optional[str] = None
    sock: typing.Optional[socket.socket] = None
    shutdown_timeout: float = 60.0
    ssl_context: typing.Optional[SSLContext] = None
    backlog: int = 128
    handle_signals: bool = False


class AccessLogger(aiohttp.abc.AbstractAccessLogger):  # noqa: R0903
    def log(self, request, response, time):  # noqa: R0201
        log = structlog.get_logger()
        log.info(
            "Access",
            remote=request.remote,
            method=request.method,
            content_type=request.content_type,
            rx=request.content_length,
            tx=response.content_length,
            path=request.path,
            time=time,
            status=response.status,
        )


async def prepare_app():
    context.add(
        aiohttp.web.Application(middlewares=[aiohttp.web.normalize_path_middleware()])
    )


async def prepare_client_session(teardown: plugin.Teardown):
    aiohttp_client_session = context.add(aiohttp.client.ClientSession())

    teardown.add(aiohttp_client_session.close())


async def prepare_server(load: plugin.Loader):
    await load(prepare_app)
    aiohttp_app = context.get(aiohttp.web.Application)

    config_source = context.get(config.ConfigSource)
    aiohttp_config = context.add(config_source.load(AioHttpConfig, "aiohttp"))

    yield aiohttp.web._run_app(  # noqa: W0212
        aiohttp_app, **attr.asdict(aiohttp_config), print=None
    )


async def prepare(load: plugin.Loader):
    await load(prepare_client_session)
    await load(prepare_server)



import pytest

PLUGINS_MARK = "buvar_plugins"


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers",
        f"{PLUGINS_MARK}(*plugins): mark test to run only on named environment",
    )


def pytest_runtest_setup(item):
    if PLUGINS_MARK in item.keywords and buvar_tasks.__name__ not in item.fixturenames:
        item.fixturenames.extend((buvar_stage.__name__, buvar_tasks.__name__))


@pytest.fixture
def buvar_config_source():
    from buvar import config

    config_source = config.ConfigSource()

    return config_source


@pytest.fixture
def buvar_context(buvar_config_source):
    from buvar import components

    context = components.Components()
    context.add(buvar_config_source)

    return context


@pytest.fixture
def buvar_stage(event_loop, buvar_context):
    from buvar import plugin

    stage = plugin.Stage(loop=event_loop, components=buvar_context)
    return stage


@pytest.fixture
def buvar_tasks(request, buvar_stage):
    # get plugins from mark
    plugins = next(
        (
            mark.args
            for mark in request.node.iter_markers()
            if mark.name == PLUGINS_MARK
        ),
        (),
    )

    try:
        # stage 1: bootstrap plugins
        if plugins:
            buvar_stage.load(*plugins)

        # stage 2: tasks
        yield buvar_stage.loader.tasks
    except Exception:
        raise

    finally:
        # stage 3: teardown
        buvar_stage.run_teardown()


@pytest.fixture
def buvar_plugin_context(buvar_stage, buvar_tasks):
    return buvar_stage.context


@pytest.hookimpl(hookwrapper=True)
def pytest_pyfunc_call(pyfuncitem):
    if PLUGINS_MARK in pyfuncitem.keywords:
        stage = pyfuncitem.funcargs[buvar_stage.__name__]
        pyfuncitem.obj = wrap_in_buvar_stage_context(stage.context, pyfuncitem.obj)

    yield


def wrap_in_buvar_stage_context(context, func):
    """Enable test function to run in plugin context."""
    import functools
    import contextvars

    from buvar.context import buvar_context

    ctx = contextvars.copy_context()

    @functools.wraps(func)
    def inner(**kwargs):
        def wrapper():
            buvar_context.set(context)
            return func(**kwargs)

        return ctx.run(wrapper)

    return inner


@pytest.fixture
def loop(event_loop):
    # use event loop from pytest-asyncio
    return event_loop


@pytest.fixture
def buvar_aiohttp_app(buvar_tasks, buvar_stage):
    import aiohttp.web

    app = buvar_stage.context.get(aiohttp.web.Application)
    return app



# cython: language_level=3
"""Self made components.

stacked::
    - each layer is the destination of additions, no inferior layer is affected
    - find collects all stacked namespaces, where superior names overwrite
      inferior ones
"""
import itertools

missing = object()


cdef class ComponentLookupError(Exception):
    pass


cdef class Components:
    def __init__(self, *stack):
        self._stack = list(stack) or [{}]  # always at least one map
        self.namespaces = self._stack[0]

    @property
    def stack(self):
        return self._stack

    cdef _add(self, object item, namespace=None, str name=None):
        cdef dict space

        if isinstance(item, type):
            raise ValueError("A component should be an instance.")

        if namespace is None:
            namespace = type(item)

        if isinstance(namespace, type):
            mro = namespace.__mro__
        else:
            mro = [namespace]

        # we add the whole MRO to have a fast lookup later
        for namespace in mro:
            if namespace in self.namespaces:
                self.namespaces[namespace][name] = item
            else:
                space = {name: item}
                self.namespaces[namespace] = space
        return item

    def add(self, item, namespace=None, *, name=None):
        if isinstance(item, type):
            raise ValueError("A component should be an instance.")
        return self._add(item, namespace, name=name)

    cdef _push(self, namespaces, stack):
        if namespaces is None:
            namespaces = {}

        components = self.__class__(namespaces, *itertools.chain(stack, self._stack))
        return components

    def push(self, namespaces=None, *stack):
        """Push a whole stack on top of the actual components or just a new
        bare namespaces one.
        """
        return self._push(namespaces, stack)

    cpdef pop(self):
        return self.__class__(*self._stack[1:])

    cdef dict _find(self, namespace):
        cdef dict merged = {}
        for namespaces in self._stack[::-1]:
            try:
                space = namespaces[namespace]
                merged.update(space)
            except KeyError:
                pass
        return merged

    def find(self, namespace):
        return self._find(namespace)

    cdef _get(self, namespace, name=None, default=missing):
        for namespaces in self._stack:
            try:
                item = namespaces[namespace][name]
                return item
            except KeyError:
                pass

        if default is missing:
            raise ComponentLookupError("Component not found", namespace, name, default)
        return default

    def get(self, namespace, *, name=None, default=missing):
        item = self._get(namespace, name=name, default=default)
        return item



import abc
import typing

import aiohttp.web
import attr

from buvar import context, util
import orjson

from .. import attrs


class Jsonify(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def loads(self, str):
        ...

    @abc.abstractmethod
    def dumps(self, obj):
        ...

    def default(self, obj):
        if hasattr(obj, "__json__"):
            return obj.__json__()
        elif attr.has(obj):
            return attrs.unstructure(obj)
        else:
            return obj


class OrjsonJsonify(Jsonify):
    def loads(self, str):
        return orjson.loads(str)

    @util.methdispatch
    def default(self, obj):
        return super().default(obj)

    def dumps(self, obj):
        return orjson.dumps(obj, default=self.default)


def response(
    data: typing.Any,
    status: int = 200,
    reason: typing.Optional[str] = None,
    headers: aiohttp.web_response.LooseHeaders = None,
) -> aiohttp.web.Response:
    jsonify = context.get(Jsonify)
    body = jsonify.dumps(data)
    return aiohttp.web.Response(
        body=body,
        status=status,
        reason=reason,
        headers=headers,
        content_type="application/json",
    )


async def prepare():
    context.add(OrjsonJsonify())



"""Very simple plugin architecture for asyncio.

`Loader` calls prepare coroutines, which yield tasks, the `Loader` collects.
These tasks are then run until complete.

A py:obj:`buvar.components.Components` context is stacked in way, that plugins
share the same context, while tasks don't, but may access the plugin context.


    >>> loop = asyncio.get_event_loop()
    >>> state = {}
    >>> async def prepare(load: Loader):
    ...     async def some_task():
    ...         state['task'] = True
    ...         return "foo"
    ...     yield some_task()

    >>> mystage = Stage()
    >>> mystage.load(prepare)
    >>> mystage.run()
    ['foo']
    >>> assert state == {'task': True}

    >>> state = {}
    >>> stage(prepare)
    ['foo']
    >>> assert state == {'task': True}
"""
import asyncio
import collections.abc
import inspect
import itertools
import sys
import types
import typing

import structlog

from . import context, util

PLUGIN_FUNCTION_NAME = "prepare"


sl = structlog.get_logger()


class Cancel(asyncio.Event):
    pass


class Teardown:
    """A collection of teardown tasks."""

    def __init__(self):
        self.tasks = []

    def add(self, task):
        self.tasks.append(task)

    def __iter__(self):
        """We teardown in the reverse order of registration."""
        return reversed(self.tasks)

    async def wait(self):
        await asyncio.gather(*self)


class Loader:
    """Load plugins and collect tasks."""

    def __init__(self):
        self._tasks = {}

    @property
    def tasks(self):
        return iter(itertools.chain(*self._tasks.values()))

    async def __call__(self, *plugins):
        """Hook the plugin from another plugin.

        The plugin may return an awaitable, an iterable of awaitable or an
        asyncgenerator of awaitables.

        :param plugins: the plugin to load
        :type plugins: list of callables which may have one argument
        """
        for plugin in plugins:
            plugin = resolve_plugin_func(plugin, caller=sys._getframe(1))
            if plugin not in self._tasks:
                # mark plugin as loaded for recursive circular stuff
                self._tasks[plugin] = []
                # load plugin
                args = collect_plugin_args(plugin)
                try:
                    result = plugin(**args)
                except TypeError as ex:
                    raise TypeError(*ex.args, plugin.__module__, plugin.__name__)
                callables = [fun async for fun in generate_async_result(result)]
                self._tasks[plugin].extend(callables)


class Stage:
    def __init__(self, components=None, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.context = (
            context.current_context()
            .push(*(components.stack if components else ()))
            .push()
        )
        # provide basic components
        self.cancel = self.context.add(Cancel(loop=self.loop))
        self.teardown = self.context.add(Teardown())
        self.loader = self.context.add(Loader())

        self.context = self.context.push()

    def load(self, *plugins):
        @context.run_child(self.context)
        def _load():
            self.loop.run_until_complete(self.loader(*plugins))

        _load()

    def run_tasks(self):
        @context.run_child(self.context)
        def _run_tasks():
            return self.loop.run_until_complete(
                run(tasks=self.loader.tasks, evt_cancel=self.cancel)
            )

        return _run_tasks()

    def run_teardown(self):
        self.loop.run_until_complete(self.teardown.wait())

    def run(self, *plugins):
        try:
            # stage 1: bootstrap plugins
            self.load(*plugins)
            # stage 2: run main task and collect teardown tasks
            return self.run_tasks()
        finally:
            # stage 3: teardown
            self.run_teardown()


def stage(*plugins, components=None, loop=None):
    """Start the asyncio process by boostrapping the root plugin.

    We have a three phase setup:

    1. run plugin hooks, to register arbitrary stuff and mainly gather a set of
    asyncio tasks for the second phase

    2. run all registered tasks together until complete
    """

    if loop is None:
        loop = asyncio.get_event_loop()

    stage = Stage(components=components, loop=loop)
    return stage.run(*plugins)


async def run(tasks, *, evt_cancel=None):
    if evt_cancel is None:
        evt_cancel = context.add(Cancel())

    # we elevate context stack for tasks
    factory = context.set_task_factory()
    try:
        fut_tasks = asyncio.gather(*map(asyncio.create_task, tasks))
    finally:
        factory.reset()

    # stop staging if we finish in any way
    fut_tasks.add_done_callback(lambda _: evt_cancel.set())

    # wait for exit
    await evt_cancel.wait()

    if not fut_tasks.done():
        # we were explicitelly stopped by cancel event
        fut_tasks.cancel()
        try:
            await fut_tasks
        except asyncio.CancelledError:
            # silence our shutdown
            pass
    else:
        return fut_tasks.result()


def collect_plugin_args(plugin):
    hints = typing.get_type_hints(plugin)
    args = {name: context.get(cls) for name, cls in hints.items()}
    return args


def resolve_plugin_func(
    plugin,
    *,
    function_name=PLUGIN_FUNCTION_NAME,
    caller: typing.Union[types.FrameType, int] = 0,
):
    plugin = util.resolve_dotted_name(
        plugin, caller=(caller + 1) if isinstance(caller, int) else caller
    )
    if inspect.ismodule(plugin):
        # apply default name
        plugin = getattr(plugin, function_name)

    if not (inspect.iscoroutinefunction(plugin) or inspect.isasyncgenfunction(plugin)):
        raise ValueError(f"{plugin} must a coroutine or an async generator.")

    return plugin


async def generate_async_result(fun_result):
    """Create an async generator out of any result."""
    if inspect.isasyncgen(fun_result):
        async for item in fun_result:
            yield item
    else:
        result = await fun_result
        if isinstance(result, collections.abc.Iterable):
            for item in result:
                yield item
        elif result is not None:
            yield result



# cython: language_level=3
cdef class ComponentLookupError(Exception):
    pass

cdef class Components:
    cdef dict namespaces
    cdef list _stack

    cdef _push(self, namespaces, stack)
    cpdef pop(self)
    cdef _add(self, item, namespace=*, str name=*)
    cdef dict _find(self, namespace)
    cdef _get(self, namespace, name=*, default=*)



"""Dependency injection.

Register a class, function or method to adapt arguments into something else.

>>> class Bar:
...     pass

>>> class Foo:
...     def __init__(self, bar: Bar):
...         self.bar = bar
...     @classmethod
...     def adapt(cls, bar: Bar) -> 'Foo':
...         return cls(bar)

>>> def adapt() -> Bar:
...     return Bar()

>>> adapters = Adapters()
>>> adapters.register(Bar)
>>> adapters.register(Foo.adapt)
>>> adapters.register(adapt)

>>> import asyncio
>>> loop = asyncio.get_event_loop()
>>> foo = loop.run_until_complete(adapters.nject(Foo))
>>> assert isinstance(foo, Foo)

"""
try:
    # gains over 100% speed up
    from .c_di import AdaptersImpl
except ImportError:
    from .py_di import AdaptersImpl


import contextvars
import collections
import inspect
import itertools
import sys
import types
import typing

import typing_inspect

from buvar import util

from .exc import ResolveError, missing


class AdapterError(Exception):
    pass


class SignatureMeta(type):
    classes = {}

    def __new__(mcs, name, bases, dct):
        cls = type.__new__(mcs, name, bases, dct)
        mcs.classes[name] = cls
        return cls

    def __call__(cls, impl, *, frame=None, **kwargs):
        # try classes from latest to oldest
        for adapter_cls in reversed(list(cls.classes.values())):
            args = adapter_cls.test_impl(impl, frame=frame)
            if args:
                inst = type.__call__(adapter_cls, impl, **args)
                return inst
        raise TypeError("No adapter implementation found", impl)


class Adapter(metaclass=SignatureMeta):
    def __init__(self, impl, *, spec, hints):
        self.impl = impl
        self.spec = spec
        self.hints = hints
        self.args = self._create_args()

    @classmethod
    def test_impl(cls, impl, *, frame):
        # TODO test for sane hints
        if callable(impl):
            return get_annotations(impl, frame=frame)

    @util.reify
    def target(self):
        return self.hints["return"]

    def register(self, adapters):
        if self.lookup not in adapters.lookups:
            adapters["index"] = collections.defaultdict(list)
            adapters.lookups[self.lookup] = True
        adapters["index"][self.target].insert(0, self)

    @staticmethod
    def lookup(adapters, target):
        if target in adapters["index"]:
            yield from adapters["index"][target]

    def _filter_arg_name(self, name):
        return False

    def _create_args(self):
        hints = self.hints
        defaults = collect_defaults(self.spec)
        args = [
            (name, extract_optional_type(hints[name]), defaults.get(name, missing))
            for name in itertools.chain(self.spec.args, self.spec.kwonlyargs)
            if not self._filter_arg_name(name)
        ]

        return args

    async def create(self, target, *args, **kwargs):
        impl = self.impl
        call = impl(*args, **kwargs)
        return await call if inspect.iscoroutinefunction(impl) else call


class ClassAdapter(Adapter):
    @util.reify
    def target(self):
        return self.impl

    @classmethod
    def test_impl(cls, impl, *, frame):
        if isinstance(impl, type):
            return super().test_impl(impl.__init__, frame=frame)

    def _filter_arg_name(self, name):
        return name is self.spec.args[0]


class ClassmethodAdapter(Adapter):
    @classmethod
    def test_impl(cls, impl, *, frame):
        if isinstance(impl, types.MethodType) and isinstance(impl.__self__, type):
            return super().test_impl(impl.__func__, frame=frame)

    def _filter_arg_name(self, name):
        return name is self.spec.args[0]


class GenericFactoryAdapter(ClassmethodAdapter):
    def __init__(self, impl, *, spec, hints, cls_type, bound):
        super().__init__(impl, spec=spec, hints=hints)
        self.cls_type = cls_type
        self.bound = bound

    @classmethod
    def test_impl(cls, impl, *, frame):
        args = super().test_impl(impl, frame=frame)
        if not args:
            return
        cls_attr = args.spec.args[0]
        generic_type = args.hints.get(cls_attr)
        # https://mypy.readthedocs.io/en/latest/generics.html#generic-methods-and-generic-self
        if (
            generic_type
            and generic_type.__origin__ is type
            and typing_inspect.is_generic_type(generic_type)
        ):
            cls_type, = typing_inspect.get_args(generic_type)
            target = args.hints["return"]

            # assure bound type equals classmethod owner
            if cls_type is target:
                bound = typing_inspect.get_bound(cls_type)
                if bound:
                    bound = bound._evaluate(frame.f_globals, frame.f_locals)
                    if bound is impl.__self__:
                        args.cls_type = cls_type
                        args.bound = bound
                        return args

    async def create(self, target, *args, **kwargs):
        impl = self.impl.__func__
        call = impl(target, *args, **kwargs)
        return await call if inspect.iscoroutinefunction(impl) else call

    def register(self, adapters):
        if self.lookup not in adapters.lookups:
            adapters["generic_index"] = collections.defaultdict(list)
            adapters.lookups[self.lookup] = True
        adapters["generic_index"][self.bound].insert(0, self)

    @staticmethod
    def lookup(adapters, target):
        # find generic adapters
        for base in inspect.getmro(target):
            if base in adapters["generic_index"]:
                yield from adapters["generic_index"][base]


class Adapters(dict, AdaptersImpl):
    def __init__(self):
        super().__init__()
        self.lookups = {}

    def register(self, *impls, frame=None):
        frame = frame or sys._getframe(1)
        for impl in impls:
            adapter = Adapter(impl, frame=frame)
            adapter.register(self)


def get_annotations(impl, *, frame=None):
    args = util.adict()
    args.hints = typing.get_type_hints(
        impl, frame.f_globals if frame else None, frame.f_locals if frame else None
    )
    args.spec = inspect.getfullargspec(impl)
    return args


def extract_optional_type(hint):
    if typing_inspect.is_optional_type(hint):
        none = type(None)
        for h in hint.__args__:
            if h is not none:
                return h
    return hint


def collect_defaults(spec):
    defaults = {
        k: v
        for k, v in itertools.chain(
            zip(reversed(spec.args or []), reversed(spec.defaults or [])),
            (spec.kwonlydefaults or {}).items(),
        )
    }
    return defaults


# our default adapters
buvar_adapters = contextvars.ContextVar(__name__)
buvar_adapters.set(Adapters())


def register(*impls):
    adapters = buvar_adapters.get()
    adapters.register(*impls, frame=sys._getframe(1))


async def nject(*targets, **dependencies):
    adapters = buvar_adapters.get()
    return await adapters.nject(*targets, **dependencies)



"""API first initiative.

At the moment this is a poor man's approach:

- only operations are mapped to routes


TODO

- schema validation
- auth
"""
import importlib.resources
import string
import typing
import urllib.parse

import aiofile
import aiohttp.web
import attr
import cached_property
import prance
import structlog

from buvar import config, di, plugin

from . import json

sl = structlog.get_logger()


@attr.s(auto_attribs=True)
class OpenApiConfig(config.Config, section="openapi"):
    path: str

    @cached_property.cached_property
    async def spec(self):
        module, _, path = self.path.partition(":")
        with importlib.resources.path(module, path) as path:
            async with aiofile.AIOFile(path) as afp:
                parser = prance.ResolvingParser(spec_string=await afp.read())
                return parser.specification


@attr.s(auto_attribs=True)
class Path:
    url: str
    parameters: typing.Optional[typing.List[typing.Dict]]


@attr.s(auto_attribs=True)
class Operation:
    id: str
    path: Path
    method: str
    responses: typing.Any
    request_body: typing.Any

    def add(self, routes: aiohttp.web.RouteTableDef, func):
        method = getattr(routes, self.method)
        method(self.path.url, name=self.id)(func)


def resolve_operations(openapi_spec):
    for url, methods in openapi_spec["paths"].items():
        parameters = methods.get("parameters")
        path = Path(url=url, parameters=parameters)
        for method, op in methods.items():
            if "parameters" == method:
                continue
            if "operationId" not in op:
                sl.warn("Operation lacks operationId", path=path, operation=op)
                continue
            operation = Operation(
                id=op["operationId"],
                path=path,
                method=method.lower(),
                responses=op.get("responses"),
                request_body=op.get("requestBody"),
            )
            sl.info("Operation", operation=operation)
            yield operation


def get_api_base(openapi_spec):
    for server in openapi_spec["servers"]:
        url = urllib.parse.urlparse(server["url"])
        return url.path


class OperationMapIncomplete(Exception):
    ...


class OperationMap:
    handlers: typing.Dict

    def __init__(self, *, allow_incomplete=False):
        self.handlers = {}
        self.allow_incomplete = allow_incomplete

    def register(self, operation_id=None):
        def _reg(func):
            op_id = operation_id if operation_id is not None else func.__name__
            if op_id in self.handlers:
                sl.warn(
                    "Operation already mapped",
                    operation_id=op_id,
                    func=self.handlers[op_id],
                )
            self.handlers[op_id] = func
            return func

        return _reg

    def create_subapp(self, spec, *, app=None):
        if app is None:
            app = aiohttp.web.Application()

        routes = aiohttp.web.RouteTableDef()
        # create subapp or prepend servers
        operations = {op.id: op for op in resolve_operations(spec)}
        for operation_id, func in self.handlers.items():
            try:
                operation = operations.pop(operation_id)
            except IndexError:
                sl.info("Operation not in API", operation=operation)
            else:
                operation.add(routes, func)

        app.add_routes(routes)
        if operations:
            if self.allow_incomplete:
                sl.warn("Not all operations mapped", unmapped=operations)
            else:
                sl.error("Not all operations mapped", unmapped=operations)
                raise OperationMapIncomplete(operations)
        return app

    async def mount(
        self,
        *,
        spec: typing.Optional[typing.Dict] = None,
        app: typing.Optional[aiohttp.web.Application] = None,
        ui: typing.Union[str, bool] = False,
    ):

        if spec is None:
            config = await di.nject(OpenApiConfig)
            spec = await config.spec

        if app is None:
            app = await di.nject(aiohttp.web.Application)

        subapp = self.create_subapp(spec)
        app.add_subapp(get_api_base(spec), subapp)

        if ui:
            openapi_routes = OpenApiRouteTableDef()
            openapi_routes.get("/spec", name=BUVAR_OPENAPI_SPEC_ROUTE)(get_openapi_spec)
            openapi_routes.get("/", name=BUVAR_OPENAPI_ROUTE)(get_openapi)
            openapi_app = OpenApiApplication()
            openapi_app.add_routes(openapi_routes)
            app.add_subapp(BUVAR_OPENAPI_PATH if ui is True else ui, openapi_app)


BUVAR_OPENAPI_PATH = "/openapi"
BUVAR_OPENAPI_SPEC_ROUTE = "buvar_openapi_spec"
BUVAR_OPENAPI_ROUTE = "buvar_openapi"


async def get_openapi_spec(request):
    config = await di.nject(OpenApiConfig)
    return json.response(await config.spec)


async def get_openapi(request):
    with importlib.resources.path("buvar.plugins.aiohttp", "openapi.html") as path:
        async with aiofile.AIOFile(path) as afp:
            text = string.Template(await afp.read()).substitute(
                openapi_spec=request.app.router[BUVAR_OPENAPI_SPEC_ROUTE].url_for()
            )
            return aiohttp.web.Response(text=text, content_type="text/html")


class OpenApiRouteTableDef(aiohttp.web.RouteTableDef):
    ...


class OpenApiApplication(aiohttp.web.Application):
    ...


async def prepare(load: plugin.Loader):
    # just load aiohttp
    await load(".")



import functools

import attr
import cattr
import typing_inspect

import pendulum


class Converter(cattr.Converter):
    """This py:obj:`Converter` is relaxed in terms of ignoring undefined and
       defaulting to None on optional attributes.
    """

    def structure_attrs_fromdict(self, obj, cl):
        # type: (Mapping, Type) -> Any
        """Instantiate an attrs class from a mapping (dict)."""
        # For public use.
        # conv_obj = obj.copy()  # Dict of converted parameters.
        conv_obj = {}
        dispatch = self._structure_func.dispatch
        for a in attr.fields(cl):
            # We detect the type by metadata.
            type_ = a.type
            if type_ is None:
                # No type.
                continue
            name = a.name
            try:
                val = obj[name]
            except KeyError:
                continue
                if typing_inspect.is_optional_type(type_):
                    if a.default is attr.NOTHING:
                        val = None
                    else:
                        continue
                else:
                    continue

            conv_obj[name] = dispatch(type_)(val, type_)

        return cl(**conv_obj)


converter = Converter()

structure = converter.structure
unstructure = converter.unstructure


def _structure_pendulum(d, _=None):
    return d if isinstance(d, pendulum.DateTime) else pendulum.parse(d, tz="UTC")


pendulum_val = functools.partial(attr.ib, converter=_structure_pendulum)
converter.register_structure_hook(pendulum.DateTime, _structure_pendulum)
converter.register_unstructure_hook(pendulum.DateTime, str)
